// ----- Change log -----
// 02/26/05 - Created
// ----------------------



// ----- Headers -----

#include "Main.h"
#include "Image.h"
#include "MapData.h"
#include <stdio.h>



// ----- Structures -----

typedef struct _RGBQUAD
{
        u8		rgbBlue;
        u8		rgbGreen;
        u8		rgbRed;
        u8		rgbReserved;

} RGBQUAD;

typedef struct _BITMAPINFOHEADER
{
        u32		biSize;
        s32		biWidth;
        s32		biHeight;
        u16		biPlanes;
        u16		biBitCount;
        u32		biCompression;
        u32		biSizeImage;
        s32		biXPelsPerMeter;
        s32		biYPelsPerMeter;
        u32		biClrUsed;
        u32		biClrImportant;

} BITMAPINFOHEADER;

typedef struct _BITMAPFILEHEADER
{
        u16		bfType;
        u32		bfSize;
        u16		bfReserved1;
        u16		bfReserved2;
        u32		bfOffBits;

} BITMAPFILEHEADER;



// ----- Global varaibles -----

Image gScreen;



// ----- Local function prototypes -----

unsigned int fgeti(size_t len, FILE *fp);



// ----- Global functions -----

void ImgCreate (Image *image, s32 width, s32 height, s32 paletteEntries)
{
	memClear (image, sizeof(Image));

	image->width = width;
	image->pitch = image->width;
	image->height = height;

	image->pixels = malloc (image->pitch*image->height);
	if (image->pixels == NULL)
	{
		ImgFree (image);
		return;
	}
	memClear (image->pixels, image->pitch*image->height);

	if (paletteEntries > 0)
	{
		image->palette = malloc (paletteEntries*sizeof(RGB));
		if (image->palette == NULL)
		{
			ImgFree (image);
			return;
		}
		memClear (image->palette, paletteEntries*sizeof(RGB));
	}
	else
	{
		image->palette = NULL;
	}
}

void ImgCreateFromBmp (Image *image, const char *fileName, BOOL loadPalette)
{
	s32 i;
	FILE *bmpFile;
	BITMAPFILEHEADER bmfh;
	BITMAPINFOHEADER bmih;

	bmpFile = fopen (fileName, "rb");
	if (bmpFile == NULL)
		return;

	bmfh.bfType				= fgeti(2, bmpFile);
	if(bmfh.bfType != 19778)
	{
		fclose (bmpFile);
		return;
	}
	bmfh.bfSize				= fgeti(4, bmpFile);
	bmfh.bfReserved1		= fgeti(2, bmpFile);
	bmfh.bfReserved2		= fgeti(2, bmpFile);
	bmfh.bfOffBits			= fgeti(4, bmpFile);

	bmih.biSize				= fgeti(4, bmpFile);
	bmih.biWidth			= fgeti(4, bmpFile);
	bmih.biHeight			= fgeti(4, bmpFile);
	bmih.biPlanes			= fgeti(2, bmpFile);
	bmih.biBitCount			= fgeti(2, bmpFile);
	bmih.biCompression		= fgeti(4, bmpFile);
	bmih.biSizeImage		= fgeti(4, bmpFile);
	bmih.biXPelsPerMeter	= fgeti(4, bmpFile);
	bmih.biYPelsPerMeter	= fgeti(4, bmpFile);
	bmih.biClrUsed			= fgeti(4, bmpFile);
	bmih.biClrImportant		= fgeti(4, bmpFile);
	if(bmih.biBitCount != 8)
	{
		fclose (bmpFile);
		return;
	}

	ImgCreate (image, bmih.biWidth, bmih.biHeight, loadPalette ? 256 : 0);

	if (loadPalette)
	{
		for(i = 0; i < 256; i++)
		{
			image->palette[i].b	= fgeti(1, bmpFile) >> 2;
			image->palette[i].g	= fgeti(1, bmpFile) >> 2;
			image->palette[i].r	= fgeti(1, bmpFile) >> 2;
			fgeti(1, bmpFile); // reserved byte
		}
	}
	else
	{
		fseek (bmpFile, 256*sizeof(RGBQUAD), SEEK_CUR);
	}

	for (i = bmih.biHeight-1; i >= 0; i--)
	{
		fread (	image->pixels + i*bmih.biWidth, 
				bmih.biWidth, 
				1, 
				bmpFile);
	}

	fclose (bmpFile);
}

void ImgFree (Image *image)
{
	if (image->pixels)
		free (image->pixels);
	if (image->palette)
		free (image->palette);

	memClear (image, sizeof(Image));
}

void ImgBlit (Image *destImg, const Image *srcImg, 
			  s32 destX, s32 destY, 
			  s32 srcX, s32 srcY, 
			  s32 width, s32 height, 
			  u32 flags)
{
	s32 i, j;
	u8 *dest, *src;

	if (destX >= destImg->width || 
		destY >= destImg->height || 
		destX + width < 0 || 
		destY + height < 0 || 
		srcX >= srcImg->width || 
		srcY >= srcImg->height || 
		srcX + width < 0 || 
		srcY + height < 0)
	{
		return;
	}

	if (destX < 0)
	{
		if (!(flags & IMAGE_BLIT_FLAG_HFLIP))
			srcX -= destX;
		width += destX;
		destX = 0;
	}
	else if (destX + width >= destImg->width)
	{
		if (flags & IMAGE_BLIT_FLAG_HFLIP)
			srcX += destX + width - destImg->width;
		width = destImg->width - destX;
	}

	if (destY < 0)
	{
		srcY -= destY;
		height += destY;
		destY = 0;
	}
	else if (destY + height >= destImg->height)
	{
		height = destImg->height - destY;
	}

	if (srcX < 0)
	{
		destX -= srcX;
		width += srcX;
		srcX = 0;
	}
	else if (srcX + width >= srcImg->width)
	{
		width = srcImg->width - srcX;
	}

	if (srcY < 0)
	{
		destY -= srcY;
		height += srcY;
		srcY = 0;
	}
	else if (srcY + height >= srcImg->height)
	{
		height = srcImg->height - srcY;
	}

	dest = destImg->pixels + destY*destImg->pitch + destX;
	src = srcImg->pixels + srcY*srcImg->pitch + srcX;

	for (j = 0; j < height; j++)
	{
		if (flags & IMAGE_BLIT_FLAG_HFLIP)
		{
			u8 *flipSrc = src + width - 1;
			for (i = 0; i < width; i++)
			{
				if (flipSrc[-i] != 0)
					dest[i] = flipSrc[-i];
			}
		}
		else
		{
			for (i = 0; i < width; i++)
			{
				if (src[i] != 0)
					dest[i] = src[i];
			}
		}
		dest += destImg->pitch;
		src += srcImg->pitch;
	}
}

void ImgRectFill (Image *image, s32 x, s32 y, s32 w, s32 h, u8 color)
{
	s32 i, j;
	u8 *dest;

	dest = image->pixels + y*image->pitch + x;

	for (j = 0; j < h; j++)
	{
		for (i = 0; i < w; i++)
		{
			dest[i] = color;
		}
		dest += image->pitch;
	}
}



// ----- Local functions -----

/* fgeti() *****************************
   Reads a little-endian (Intel-style) integer of len bytes from
   the given file fp and returns it.  On EOF, sets all high order
   bits to 0 and sets feof(fp).
*/
unsigned int fgeti(size_t len, FILE *fp)
{
  unsigned int ret = 0, bit_offset = 0;

  for(; len > 0; len--)
  {
    unsigned int c = fgetc(fp);
    if(c == EOF)
      return ret;
    if(bit_offset < 32)
    {
      ret |= (c & 0xff) << bit_offset;
      bit_offset += 8;
    }
  }
  return ret;
}
