#include "Main.h"
#include "Font.h"
#include "Game.h"
#include "Map.h"
#include "MapData.h"
#include "TextBox.h"



// ----- Global functions -----

void MapInit (u32 level, Map *map)
{
	s32 i;
	const MapFile *mapFile;
	Image mapImg;

	memClear (map, sizeof (Map));

	mapFile = &dMap[level];

	map->mode = MAP_MODE_GAME;

	ImgCreateFromBmp (&map->imgLifeMeter, "Lives.bmp", FALSE);

		// Main map layer
	ImgCreateFromBmp (&mapImg, mapFile->mapBmpFile, 0);

	map->width		= mapImg.width;
	map->height		= mapImg.height;
	map->mapTile	= malloc (map->width*map->height);
	memcpy (map->mapTile, mapImg.pixels, map->width*map->height);

	ImgFree (&mapImg);

		// BG layer
	ImgCreateFromBmp (&mapImg, mapFile->mapBgBmpFile, 0);

	map->bgWidth		= mapImg.width;
	map->bgHeight		= mapImg.height;
	map->xScrollMin		= mapFile->xScrollMin;
	map->yScrollMin		= mapFile->yScrollMin;
	map->xScrollMax		= mapFile->xScrollMax;
	map->yScrollMax		= mapFile->yScrollMax;
	map->bgXScrollMul	= mapFile->bgXScrollMul;
	map->bgYScrollMul	= mapFile->bgYScrollMul;
	map->bgXScrollOfs	= mapFile->bgXScrollOfs;
	map->bgYScrollOfs	= mapFile->bgYScrollOfs;
	map->mapBgTile		= malloc (map->bgWidth*map->bgHeight);
	memcpy (map->mapBgTile, mapImg.pixels, map->bgWidth*map->bgHeight);

	ImgFree (&mapImg);

		// Tileset
	ImgCreateFromBmp (&map->tileGfx, mapFile->tileBmpFile, TRUE);
	memcpy (gScreen.palette, map->tileGfx.palette, 256*sizeof(RGB));
	gScreen.palette[OIRAM_PAL_OVERALLS] = dOiramOverallsNormal;
	map->tileCollision	= mapFile->tileCollision;

		// Sprites
	for (i = 0; i < mapFile->spriteCount; i++)
	{
		Sprite *spt = &map->sprite[i];
		const SptListFile *listFile = &mapFile->spriteList[i];

		SptCreate (spt, listFile, map);
	}

	if (mapFile->followSprite != 0xff)
	{
		map->followSprite = &map->sprite[mapFile->followSprite];
	}
}

void MapExit (Map *map)
{
	memClear (gScreen.palette, 256*sizeof(RGB));
	memClear (gScreen.pixels, gScreen.pitch*gScreen.height);
	if (map->mapTile)
		free (map->mapTile);
	ImgFree (&map->tileGfx);
	ImgFree (&map->imgLifeMeter);
}

void MapUpdate (Map *map)
{
	s32 i, j, k, l;
	Sprite *collisionList[MAP_MAX_SPRITES];
	s32 collisionNum = 0;

	switch (map->mode)
	{
	case MAP_MODE_GAME:
		for (i = 0; i < MAP_MAX_SPRITES; i++)
		{
			Sprite *spt = &map->sprite[i];

			if (spt->flags & SPRITE_FLAG_BORN)
			{
				spt->flags &= ~SPRITE_FLAG_BORN;
				spt->flags |= SPRITE_FLAG_ALIVE;
			}

			if (spt->flags & SPRITE_FLAG_ALIVE)
			{
				collisionList[collisionNum++] = spt;
			}
		}

		for (i = 0; i < collisionNum; i++)
		{
			Sprite *spt = collisionList[i];
			const SptAnim *anim = spt->anim[spt->curAnim];

			if (spt->update != NULL)
			{
				spt->update (spt, map);
			}

			if (anim->frame[spt->frame].delay != 0)
			{
				if (--spt->delay == 0)
				{
					if (++spt->frame >= anim->frameCount)
					{
						spt->frame = 0;
					}

					spt->delay = anim->frame[spt->frame].delay;
				}
			}

			spt->newX = spt->x + spt->xVel;
			spt->newY = spt->y + spt->yVel;
		}

		for (i = 0; i < collisionNum; i++)
		{
			Sprite *spt = collisionList[i];
			s32 sLeft, sRight, sTopH, sBottomH, sTopV, sBottomV;

			if (spt->flags & SPRITE_FLAG_COLLIDE)
			{
				sLeft	 = spt->newX + SPT_FX(spt->collision.x);
				sRight	 = sLeft + SPT_FX(spt->collision.w);
				sTopH	 = spt->y + SPT_FX(spt->collision.y);
				sBottomH = sTopH + SPT_FX(spt->collision.h);

				sTopV	 = spt->newY + SPT_FX(spt->collision.y);
				sBottomV = sTopV + SPT_FX(spt->collision.h);



				// --- Handle map collision ---

				{
					s32 tileStartX, tileStartY, tileWidth, tileHeight;
					s32 sTop, sBottom;
					u32 dir;

					sTop		= sTopH;
					sBottom		= sBottomH;

					if (spt->newX < spt->x)
						dir = SPRITE_COL_DIR_LEFT;
					else
						dir = SPRITE_COL_DIR_RIGHT;

					tileStartX	= SPT_FX_TO_INT(sLeft) / TILE_SIZE;
					tileStartY	= SPT_FX_TO_INT(sTopH) / TILE_SIZE;
					tileWidth	= ((spt->collision.w + TILE_SIZE-1) / TILE_SIZE) + 1;
					tileHeight	= ((spt->collision.h + TILE_SIZE-1) / TILE_SIZE) + 1;

					for (l = 0; l < 2; l++)
					{
						for (k = 0; k < tileHeight; k++)
						{
							for (j = 0; j < tileWidth; j++)
							{
								s32 tLeft, tRight, tTop, tBottom;
								s32 tx, ty;
								u8 tileCol;

								tx = tileStartX+j;
								ty = tileStartY+k;

								if ((tx >= 0 && tx < map->width) && 
									(ty >= 0 && ty < map->height))
								{
									tileCol = map->tileCollision[map->mapTile[ty*map->width + tx]];

									if ((spt->flags & SPRITE_FLAG_COLLIDE) && 
										(tileCol & (1<<(SPRITE_COL_DIR_MAX-dir))))
									{
										tLeft	= SPT_FX(tx*TILE_SIZE);
										tTop	= SPT_FX(ty*TILE_SIZE);
										tRight	= tLeft + SPT_FX(TILE_SIZE);
										tBottom	= tTop  + SPT_FX(TILE_SIZE);

										if (!(sLeft    >= tRight  || 
											  sRight   <= tLeft   || 
											  sTop	   >= tBottom || 
											  sBottom  <= tTop))
										{
											if (spt->colHandler != NULL)
											{
												spt->colHandler (spt, NULL, dir | (tileCol & 0xf0), map);

												sLeft	 = spt->newX + SPT_FX(spt->collision.x);
												sRight	 = sLeft + SPT_FX(spt->collision.w);
												sTopV	 = spt->newY + SPT_FX(spt->collision.y);
												sBottomV = sTopV + SPT_FX(spt->collision.h);
											}
										}
									}
								}
							}
						}

						if (spt->newY < spt->y)
							dir = SPRITE_COL_DIR_UP;
						else
							dir = SPRITE_COL_DIR_DOWN;

						sTop		= sTopV;
						sBottom		= sBottomV;

						tileStartX	= SPT_FX_TO_INT(sLeft) / TILE_SIZE;
						tileStartY	= SPT_FX_TO_INT(sTopV) / TILE_SIZE;

					}	// end H/V test loop
				}	// end map collision



				// --- Handle sprite collision ---

				for (j = i + 1; j < collisionNum; j++)
				{
					Sprite *obs = collisionList[j];
					s32 oLeft, oRight, oTop, oBottom;
					BOOL collided;

					collided = FALSE;

					if ((spt->flags & SPRITE_FLAG_COLLIDE) && 
						(obs->flags & SPRITE_FLAG_COLLIDE))
					{
						// - Horizontal movement -

						oLeft	= obs->newX + SPT_FX(obs->collision.x);
						oRight	= oLeft + SPT_FX(obs->collision.w);
						oTop	= obs->y + SPT_FX(obs->collision.y);
						oBottom	= oTop + SPT_FX(obs->collision.h);

						if (!(sLeft    >= oRight  || 
							  sRight   <= oLeft   || 
							  sTopH	   >= oBottom || 
							  sBottomH <= oTop))
						{
							u32 dir;

							collided = TRUE;

							if (spt->newX < spt->x)
								dir = SPRITE_COL_DIR_LEFT;
							else
								dir = SPRITE_COL_DIR_RIGHT;

							if (spt->colHandler != NULL)
							{
								spt->colHandler (spt, obs, dir, map);

								sLeft	 = spt->newX + SPT_FX(spt->collision.x);
								sRight	 = sLeft + SPT_FX(spt->collision.w);
								sTopV	 = spt->newY + SPT_FX(spt->collision.y);
								sBottomV = sTopV + SPT_FX(spt->collision.h);
							}

							if (obs->colHandler != NULL)
							{
								obs->colHandler (obs, spt, SPRITE_COL_DIR_MAX - dir, map);

								oLeft	= obs->newX - obs->hotspotX + SPT_FX(obs->collision.x);
								oRight	= oLeft + SPT_FX(obs->collision.w);
							}
						}
					}


					// - Vertical movement -

					if ((collided == FALSE) && 
						(spt->flags & SPRITE_FLAG_COLLIDE) && 
						(obs->flags & SPRITE_FLAG_COLLIDE))
					{
						oTop	= obs->newY + SPT_FX(obs->collision.y);
						oBottom	= oTop + SPT_FX(obs->collision.h);

						if (!(sLeft    >= oRight  || 
							  sRight   <= oLeft   || 
							  sTopV	   >= oBottom || 
							  sBottomV <= oTop))
						{
							u32 dir;

							if (spt->newY < spt->y)
								dir = SPRITE_COL_DIR_UP;
							else
								dir = SPRITE_COL_DIR_DOWN;

							if (spt->colHandler != NULL)
							{
								spt->colHandler (spt, obs, dir, map);

								sLeft	 = spt->newX + SPT_FX(spt->collision.x);
								sRight	 = sLeft + SPT_FX(spt->collision.w);
								sTopV	 = spt->newY + SPT_FX(spt->collision.y);
								sBottomV = sTopV + SPT_FX(spt->collision.h);
							}

							if (obs->colHandler != NULL)
							{
								obs->colHandler (obs, spt, SPRITE_COL_DIR_MAX - dir, map);
							}
						}
					}
				}	// end obs loop
			}	// end spt collide flag set

			spt->x = spt->newX;
			spt->y = spt->newY;

		}	// end spt loop

		if (map->followSprite != NULL)
		{
			map->xScroll = SPT_FX_TO_INT(map->followSprite->x) - gScreen.width/2;
			map->yScroll = SPT_FX_TO_INT(map->followSprite->y) - gScreen.height/2;

			if (map->xScroll < map->xScrollMin)
				map->xScroll = map->xScrollMin;
			else if (map->xScroll + gScreen.width >= map->xScrollMax)
				map->xScroll = map->xScrollMax - gScreen.width;

			if (map->yScroll < map->yScrollMin)
				map->yScroll = map->yScrollMin;
			else if (map->yScroll + gScreen.height >= map->yScrollMax)
				map->yScroll = map->yScrollMax - gScreen.height;
		}

		break;	// end case MAP_MODE_GAME

	case MAP_MODE_TEXTBOX:
		if (map->textBoxTimer > 0)
		{
			if (--map->textBoxTimer == 0)
			{
				clear_keybuf ();
			}
		}
		else
		{
			if (keypressed ())
			{
				if (map->flags & MAP_FLAG_END_GAME)
				{
					GameSetMode (&gGame, GAME_MODE_END, 0);
				}
				else
				{
					map->mode = MAP_MODE_GAME;
				}
			}
		}
		break;

	default:
		ASSERT_ALWAYS();
	}
}

void MapRender (Map *map, Image *screen)
{
	s32 i, j, k, 
		xScroll, yScroll, 
		width, height, 
		xStart, yStart, 
		xOfs, yOfs, 
		tileWidth, tileHeight, 
		xTile, yTile, 
		sptPriStart, sptPriEnd;
	const u8 *mapTile;

	xScroll		= (map->xScroll*map->bgXScrollMul >> 8) + map->bgXScrollOfs;
	yScroll		= (map->yScroll*map->bgYScrollMul >> 8) + map->bgYScrollOfs;
	width		= map->bgWidth;
	height		= map->bgHeight;
	mapTile		= map->mapBgTile;

	sptPriStart = SPRITE_PRIORITY_MAX;
	sptPriEnd	= SPRITE_PRIORITY_2;

	tileWidth	= ((screen->width  + TILE_SIZE-1) / TILE_SIZE) + 2;
	tileHeight	= ((screen->height + TILE_SIZE-1) / TILE_SIZE) + 2;

	for (k = 0; k < 2; k++)
	{
		xStart		= xScroll / TILE_SIZE;
		yStart		= yScroll / TILE_SIZE;

		xOfs		= xScroll % TILE_SIZE;
		yOfs		= yScroll % TILE_SIZE;

		for (j = 0; j < tileHeight; j++)
		{
			for (i = 0; i < tileWidth; i++)
			{
				xTile = xStart+i;
				yTile = yStart+j;

				if (xTile >= 0 && xTile < width && 
					yTile >= 0 && yTile < height)
				{
					ImgBlit (screen, 
							 &map->tileGfx, 
							 i*TILE_SIZE - xOfs, j*TILE_SIZE - yOfs, 
							 0, mapTile[yTile*width + xTile]*TILE_SIZE, 
							 TILE_SIZE, TILE_SIZE, 0);
				}
			}
		}

		for (j = sptPriStart; j >= sptPriEnd; j--)
		{
			for (i = 0; i < MAP_MAX_SPRITES; i++)
			{
				Sprite *spt = &map->sprite[i];

				if ((spt->flags & SPRITE_FLAG_ALIVE) && 
					(spt->priority == j))
				{
					spt->render (spt, map, screen);
				}
			}
		}

		xScroll		= map->xScroll;
		yScroll		= map->yScroll;
		width		= map->width;
		height		= map->height;
		mapTile		= map->mapTile;

		sptPriStart	= SPRITE_PRIORITY_1;
		sptPriEnd	= 0;
	}

	if (map->mode == MAP_MODE_TEXTBOX)
	{
		TextBoxDraw (&map->textBox, screen);
	}

	ImgBlit (screen, &map->imgLifeMeter, 8, 8, gGame.lives*32, 0, 32, 32, 0);
}

Sprite *MapCreateSprite (Map *map, const SptListFile *listFile)
{
	s32 i;

	for (i = 0; i < MAP_MAX_SPRITES; i++)
	{
		if (!(map->sprite[i].flags & SPRITE_FLAG_ALIVE))
		{
			SptCreate (&map->sprite[i], listFile, map);
			return &map->sprite[i];
		}
	}

	return NULL;
}

void MapCreateTextBox (Map *map, const char *text)
{
	TextBoxCreate (&map->textBox, 
				   text, 
				   gScreen.width/2, 
				   gScreen.height/4, 
				   MAP_TEXTBOX_WIDTH, 
				   MAP_TEXTBOX_HEIGHT, 
				   FONT_COLORS(4, 8, 8));
	map->mode = MAP_MODE_TEXTBOX;
	map->textBoxTimer = MAP_TEXTBOX_TIME;
}



// --- Sprite functions ---

void SptCreate (Sprite *spt, const SptListFile *listFile, Map *map)
{
	const SptTypeFile *sptFile = &dSptType[listFile->spriteType];

	memClear (spt, sizeof(Sprite));

	ImgCreateFromBmp (&spt->image, sptFile->imageBmpFile, FALSE);
	spt->type		= listFile->spriteType;
	spt->flags		= sptFile->flags | SPRITE_FLAG_BORN;
	spt->width		= sptFile->width;
	spt->height		= sptFile->height;
	spt->hotspotX	= sptFile->hotspotX;
	spt->hotspotY	= sptFile->hotspotY;
	spt->anim		= sptFile->anim;
	spt->frame		= 0;
	spt->curAnim	= 0;
	spt->delay		= spt->anim[spt->curAnim]->frame[spt->frame].delay;
	spt->update		= sptFile->update;
	spt->render		= sptFile->render;
	spt->colHandler	= sptFile->colHandler;
	spt->collision	= sptFile->collision;
	spt->priority	= listFile->priority;
	spt->x			= listFile->x;
	spt->y			= listFile->y;

	sptFile->init (spt, map);
}

void SptSetAnim (Sprite *spt, u8 anim, BOOL restart)
{
	if (restart == TRUE || anim != spt->curAnim)
	{
		spt->curAnim	= anim;
		spt->frame		= 0;
		spt->delay		= spt->anim[spt->curAnim]->frame[spt->frame].delay;
	}
}

void SptRenderImg (Sprite *spt, Map *map, Image *screen)
{
	s32 xScroll, yScroll;

	if (map != NULL)
	{
		xScroll = map->xScroll;
		yScroll = map->yScroll;
	}
	else
	{
		xScroll = 0;
		yScroll = 0;
	}

	ImgBlit (screen, &spt->image, 
		SPT_FX_TO_INT (spt->x - spt->hotspotX) - xScroll, 
		SPT_FX_TO_INT (spt->y - spt->hotspotY) - yScroll, 
		spt->anim[spt->curAnim]->frame[spt->frame].frame*spt->width, 0, 
		spt->width, spt->height, 
		(spt->flags & SPRITE_FLAG_HFLIP) && (spt->type != SPRITE_TYPE_TROLL) ? IMAGE_BLIT_FLAG_HFLIP : 0);
}

	// Returns TRUE if snapped successfully, FALSE if ground too far below
BOOL SptGroundSnap (Sprite *spt, Map *map, BOOL setPos)
{
	s32 i, j;
	Rect sptRect;
	s32 tileStartX, tileStartY, tileWidth;
	BOOL retVal = FALSE;

	if (spt->flags & SPRITE_FLAG_COLLIDE)
	{
		SptGetColRect (spt, &sptRect, spt->x, spt->y + SPT_GROUND_SNAP_DIST);

		for (j = 0; j < MAP_MAX_SPRITES; j++)
		{
			Sprite *obs = &map->sprite[j];

			if ((obs->flags & SPRITE_FLAG_ALIVE) && 
				(obs->flags & SPRITE_FLAG_COLLIDE) && 
				SPRITE_IS_BLOCK (obs) && 
				(obs != spt))
			{
				Rect obsRect;

				SptGetColRect (obs, &obsRect, obs->newX, obs->newY);
				if (SptCollideRect (spt, &sptRect, &obsRect) == TRUE)
				{
					spt->newY = spt->y = obsRect.top - 
						SPT_FX(spt->collision.y + spt->collision.h);
					SptGetColRect (spt, &sptRect, spt->x, spt->y);
					retVal = TRUE;
				}
			}
		}

		tileStartX	= SPT_FX_TO_INT(sptRect.left) / TILE_SIZE;
		tileStartY	= SPT_FX_TO_INT(sptRect.bottom) / TILE_SIZE;
		tileWidth	= ((spt->collision.w + TILE_SIZE-1) / TILE_SIZE) + 1;

		for (i = 0; i < tileWidth; i++)
		{
			Rect tileRect;
			s32 tx, ty;

			tx = tileStartX+i;
			ty = tileStartY;

			if ((spt->flags & SPRITE_FLAG_COLLIDE) && 
				(tx >= 0 && tx < map->width) && 
				(ty >= 0 && ty < map->height) && 
				(map->tileCollision[map->mapTile[ty*map->width + tx]] & COLLISION_TOP))
			{
				tileRect.left	= SPT_FX(tx*TILE_SIZE);
				tileRect.top	= SPT_FX(ty*TILE_SIZE);
				tileRect.right	= tileRect.left + SPT_FX(TILE_SIZE);
				tileRect.bottom	= tileRect.top  + SPT_FX(TILE_SIZE);

				if (SptCollideRect (spt, &sptRect, &tileRect) == TRUE)
				{
					if (setPos == TRUE)
						spt->newY = spt->y = (spt->y + SPT_GROUND_SNAP_DIST) & SPT_FX(~(TILE_SIZE-1));

					return TRUE;
				}
			}
		}
	}

	return retVal;
}

void SptGetColRect (Sprite *spt, Rect *rect, s32 x, s32 y)
{
	rect->left		= x + SPT_FX(spt->collision.x);
	rect->top		= y + SPT_FX(spt->collision.y);
	rect->right		= rect->left + SPT_FX(spt->collision.w);
	rect->bottom	= rect->top  + SPT_FX(spt->collision.h);
}

BOOL SptCollideRect (Sprite *spt, Rect *sptRect, Rect *obsRect)
{
	if (!(sptRect->left    >= obsRect->right  || 
		  sptRect->right   <= obsRect->left   || 
		  sptRect->top	   >= obsRect->bottom || 
		  sptRect->bottom  <= obsRect->top))
	{
		return TRUE;
	}
	else
	{
		return FALSE;
	}
}
