// ----- Change log -----
// 02/26/05 - Created
// ----------------------

#ifndef MAP_H
#define MAP_H

#include "Main.h"
#include "Image.h"
#include "MapData.h"
#include "TextBox.h"

	// Sprite headers
#include "Block.h"
#include "Flower.h"
#include "Oiram.h"
#include "SptSine.h"
#include "Turtle.h"



// ----- Constants -----

#define TILE_SIZE		32
#define TILE_BYTES		(TILE_SIZE*TILE_SIZE)
#define MAP_MAX_SPRITES	32
#define MAP_TEXTBOX_WIDTH	300
#define MAP_TEXTBOX_HEIGHT	200
#define MAP_TEXTBOX_TIME	10

#define SPT_FX_BITS		8
#define SPT_GRAVITY		SPT_FX(5)
#define SPT_GRAVITY_MAX	SPT_FX(16)
#define SPT_GROUND_SNAP_DIST	SPT_FX(TILE_SIZE/2)

typedef enum _SPRITE_COL_DIR
{
	SPRITE_COL_DIR_LEFT,
	SPRITE_COL_DIR_UP,
	SPRITE_COL_DIR_DOWN,
	SPRITE_COL_DIR_RIGHT,

	SPRITE_COL_DIR_NUM,
	SPRITE_COL_DIR_MAX = SPRITE_COL_DIR_NUM - 1,

} SPRITE_COL_DIR;

typedef enum _COLLISION_TYPE
{
	COLLISION_LEFT		= 1<<SPRITE_COL_DIR_LEFT,
	COLLISION_TOP		= 1<<SPRITE_COL_DIR_UP,
	COLLISION_BOTTOM	= 1<<SPRITE_COL_DIR_DOWN,
	COLLISION_RIGHT		= 1<<SPRITE_COL_DIR_RIGHT,

	COLLISION_HAZARD	= 1<<4,
	COLLISION_PIPE		= 2<<4,
	COLLISION_TEXT		= 3<<4,

	COLLISION_NONE		= 0,
	COLLISION_SOLID		= 0xF,

} COLLISION_TYPE;

typedef enum _SPRITE_FLAGS
{
	SPRITE_FLAG_ALIVE	= BIT00,
	SPRITE_FLAG_BORN	= BIT01,
	SPRITE_FLAG_HFLIP	= BIT02,
	SPRITE_FLAG_COLLIDE	= BIT03,

} SPRITE_FLAGS;

typedef enum _SPRITE_PRIORITY
{
	SPRITE_PRIORITY_0,
	SPRITE_PRIORITY_1,
	SPRITE_PRIORITY_2,
	SPRITE_PRIORITY_3,

	SPRITE_PRIORITY_NUM,
	SPRITE_PRIORITY_MAX = SPRITE_PRIORITY_NUM - 1,

} SPRITE_PRIORITY;

typedef enum _MAP_FLAGS
{
	MAP_FLAG_DIE		= BIT00,
	MAP_FLAG_NEXT_LEVEL	= BIT01,
	MAP_FLAG_END_GAME	= BIT02,

} MAP_FLAGS;

typedef enum _MAP_MODE
{
	MAP_MODE_GAME,
	MAP_MODE_TEXTBOX,

	MAP_MODE_NUM,
	MAP_MODE_MAX = MAP_MODE_NUM - 1,

} MAP_MODE;



// ----- Macros -----

#define SPT_FX(x)			((x) << SPT_FX_BITS)
#define SPT_FX_TO_INT(x)	((x) >> SPT_FX_BITS)



// ----- Structures -----

typedef struct _SptVars
{
	union
	{
		SptVarsOiram	oiram;
		SptVarsTurtle	turtle;
		SptVarsBlock	block;
		SptVarsFlower	flower;
		SptVarsSine		sine;
	};

} SptVars;

typedef struct _Sprite
{
	SptUpdateFunc	update;
	SptRenderFunc	render;
	SptColHandler	colHandler;
	Image			image;			// Frames are horizontally aligned
	s8Rect			collision;		// Relative to hotspot
	s32				x, y;			// SPT_FX_BITS bit fixed-point
	s32				newX, newY;
	s32				xVel, yVel;
	s32				hotspotX, hotspotY;
	s32				width, height;
	const SptAnim	**anim;			// List of actual frame numbers
	u8				curAnim;
	u8				frame;			// Index into anim
	u8				delay;
	u8				priority;		// SPRITE_PRIORITY
	u8				type;			// SPRITE_TYPE
	u8				flags;			// SPRITE_FLAGS
	SptVars			vars;

} Sprite;

typedef struct _Map
{
	Sprite		sprite[MAP_MAX_SPRITES];
	Sprite		*followSprite;
	u8			mode;				// MAP_MODE
	u8			flags;				// MAP_FLAGS

	s32			width, height;
	s32			xScroll, yScroll;
	s32			xScrollMin, yScrollMin;
	s32			xScrollMax, yScrollMax;
	u8			*mapTile;

	s32			bgWidth, bgHeight;
	s32			bgXScrollMul, bgYScrollMul;
	s32			bgXScrollOfs, bgYScrollOfs;
	u8			*mapBgTile;

	Image		tileGfx;
	const u8	*tileCollision;		// COLLISION_TYPE

	Image		imgLifeMeter;

	TextBox		textBox;
	u8			textBoxTimer;

} Map;



// ----- Global variables -----

extern Map gMap;



// ----- Global functions -----

extern void MapInit (u32 level, Map *map);
extern void MapExit (Map *map);
extern void MapUpdate (Map *map);
extern void MapRender (Map *map, Image *screen);
extern Sprite *MapCreateSprite (Map *map, const SptListFile *listFile);
extern void MapCreateTextBox (Map *map, const char *text);

extern void SptCreate (Sprite *spt, const SptListFile *listFile, Map *map);
extern void SptSetAnim (Sprite *spt, u8 anim, BOOL restart);
extern void SptRenderImg (Sprite *spt, Map *map, Image *screen);
extern BOOL SptGroundSnap (Sprite *spt, Map *map, BOOL setPos);
extern void SptGetColRect (Sprite *spt, Rect *rect, s32 x, s32 y);
extern BOOL SptCollideRect (Sprite *spt, Rect *sptRect, Rect *obsRect);

#endif	// MAP_H